/* Krzysztof Krygiel, grupa 4I1
 * 
 * Przykad wykorzystania biblioteki boost::graph
 *
 * Poniszy program demonstruje moliwoci uycia wyej wymienionej biblioteki
 * w celu wyznaczenia najkrtszych cieek w waonym grafie przy pomocy
 * algorytmu Dijkstry ( O(V log V) ). Naley pamita, e za pomoc tego 
 * algorytmu moliwe jest wyznaczenie szukanych odlegoci tylko dla grafu 
 * o nieujemnych wagach krawdzi. Dla grafu o ujemnych wagach (ktry nie zawiera 
 * cyklu o ujemnej dugoci) naley wykorzysta algorytm Bellmana-Forda. 
 * Dla grafu o jednakowych  wartociach wag mona uy przeszukiwania w szerz 
 * w celu optymalizacji programu ( O(V+E) ).
 * 
 * Pod koniec programu pokazano jak w atwy sposb znale minimalne drzewo 
 * rozpinajce.w nieskierowanym grafie. Zaimplementowany w bibliotece algorytm
 * Prima ma zoono O(E log V) gdzie E to liczba krawdzi, a V to liczba
 * wierzchokw.
 * Zaimplementowany zosta take algorytm Kruskala o zoonoci O(E log E).
 * Ze wzgldu na mniejsz wydajno jego uycie nie zostanie tutaj przedstawione.
 */

#include <iostream>

#include <boost/lexical_cast.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/graph/prim_minimum_spanning_tree.hpp>

using namespace boost;
using namespace std;

/*
 * Klasa wizytatora dla algorytmu Dijkstry. Po przetworzeniu danego wierzchoka 
 * wypisywany jest odpowiedni komunikat.
 * Istnieje jeszcze 6 innych metod, ktre mona wykorzysta w tym wizytatorze.
 */
class Wizytator:public default_dijkstra_visitor {
	public:
		template<class Etykieta, class Graf> 
		void examine_vertex(Etykieta& w, Graf& g){		//parametrami jest etykieta wierzcholka i graf na ktrym wykonujemy algorytm
			cout<<"Wyznaczono odlego do wierzchoka "<<static_cast<char>(w+'A')<<endl;
		}
};


int main(int, char *[])
{
	//Definicja typu danych do przechowywania grafu (graf bdzie reprezentowany przez listy ssiedztwa)
	typedef adjacency_list<
		listS,			//krawdzie przechowywane w std::list
		vecS,				//wierzchoki przechowywane w std::vector
		undirectedS,		//graf nieskierowany (istnieje molwo stworzenia grafu skierowanego : directedS)
		no_property,	//Waciwoci wierzchokw (brak atrybutw)
		property<edge_weight_t, int> //Waciwoci krawdzi (kadej krawdzi przyporzdkowujemy wag)
	> graf;
	
	typedef graph_traits < graf >::vertex_descriptor vertex_descriptor;
	typedef graph_traits < graf >::edge_descriptor edge_descriptor;
	
	//Krawedzie reprezentowane s przez pare liczb oznaczajacych wierzchoek pocztkowy i kocowy
	typedef pair<int,int> Krawedz;
	//nazwy symboliczne dla wierzchokw, liczba_wierz - liczba wierzchokw w grafie
	enum wierzcholki {A,B,C,D,E,F,G, liczba_wierz};
		
	//inicjalizacja krawdzi grafu
	Krawedz krawedzie[] = { Krawedz(A, B), Krawedz(A, C), Krawedz(B, C), Krawedz(B, D), Krawedz(D, A),
		Krawedz(D, E), Krawedz(D, G), Krawedz(E, B), Krawedz(E, F), Krawedz(F, G) };
		
	//wagi poszczeglnych krawdzi (kolejno odpowiada kolejnoci w krawedzie[])
	int wagi[] = { 3, 2, 4, 2, 10, 1, 3, 1, 5, 15  };
	int liczba_kraw = sizeof(krawedzie) / sizeof(Krawedz);

	//tworzenie grafu
	graf g(
		krawedzie,			//iterator na pierwszy element konteneru zawierajcego krawdzie
		krawedzie+liczba_kraw,		//iterator na element za ostatnim w kontenerze zawierajcym krawdzie
		wagi,			//wagie krawdzi
		liczba_wierz		//liczba wierzchokw
	);

	//przygotowanie kontenerw (dla odlegoci do danego wierzchoka i poprzednika na najkrtszej sciece)
	vector<vertex_descriptor> poprzednik(liczba_wierz);
	vector<int> odleglosc(liczba_wierz);

	//wywoanie algorytmu wyszukiwania najkrtszej cieki
	dijkstra_shortest_paths(
		g,			//przeszukiwany graf
		vertex(D, g),	//wierzchoek startowy
		//informacja o miejscu zapisu wynikw
		predecessor_map(make_iterator_property_map(poprzednik.begin(), get(vertex_index, g))).
			distance_map(make_iterator_property_map(odleglosc.begin(), get(vertex_index, g)))
	);

	//wypisanie wynikw
	for (int i=0;i<liczba_wierz;i++){
		//jeeli poprzednikiem wierzchoka jest on sam (i nie jest to wierzchoek startowy) to
		//dany wierzchoek nie jest osigalny z wierzchoka startowego
		if (i!=D && poprzednik[i]==i) cout<<"Brak scieki do wierzchoka "<< static_cast<char>(i+'A')<<endl;
		else{
			//proste odtworzenie penej cieki poprzez przesuwanie si od koca po poprzednikach
			string sciezka=lexical_cast<string>(static_cast<char>(i+'A'));
			int wierz=i;
			while (poprzednik[wierz]!=wierz) {
				wierz=poprzednik[wierz];
				sciezka=lexical_cast<string>(static_cast<char>(wierz+'A'))+sciezka;
			}
			
			cout<<"Odlego do wierzchoka "<<(char)(i+'A')<<" wynosi "<<odleglosc[i]<<" (dla scieki: "<<sciezka<<")"<<endl;
		}
	}
	
	/*
	 * Oprocz prostej funkcji implementujcej algorytm Dijkstry, ktra pozwala na zdobycie informacji o 
	 * najkrtszej ciece istnieje take wzorzec umoliwiajcy doczenie wasnego wizytatora. Daje on
	 * moliwo reagowania na poszczeglne elementy algorytmu Dijkstry (np. inicjacja czy pobranie
	 * wierzchoka z kolejki priorytetowej).
	 */
	
	//wywoanie algorytmu z przekazaniem wasnego wizytatora
	property_map<graf, edge_weight_t>::type weightmap = get(edge_weight, g);
	property_map<graf, vertex_index_t>::type indexmap = get(vertex_index, g);
	
	dijkstra_shortest_paths(
		g,
		vertex(D, g),
		&poprzednik[0],
		&odleglosc[0],
		weightmap,		//wagi krawedzi pobrane z grafu
		indexmap, 		//indeksy (etykiety) krawedzi pobrane z grafu
		std::less<int>(),		//operator porwnania (dla porwnywania odlegoci)
		closed_plus<int>(),		//operator dodawania (sumowanie odlegosci)
		(std::numeric_limits<int>::max)(),		//warto, ktra zostan zaincicjowane odlegoci (dla poprawnoci musi by wiksza od najdluszej cieki)
		0,		//odlegoc do pocztkowego wierzchoka
		Wizytator()		//wlasny wizytotor (mozna przekazac domyslny -default_dijkstra_visitor()
	);
	
	
	
	//Znajdowanie minimalnego drzewa rozpinajcego dla wczeniej skonstruowanego grafu
	prim_minimum_spanning_tree(g, &poprzednik[0]);
	
	cout<<"Do minimalnego drzewa rozpinajcego nale krawdzie:"<<endl;
	for (int i=0;i<liczba_wierz;i++){
		//krawed naley do drzewa rozpinajcego jeeli poprzednik[i]!=i
		if (poprzednik[i]!=i) cout<<"("<<static_cast<char>(i+'A')<<" "<<static_cast<char>(poprzednik[i]+'A')<<")"<<endl;
	}
}

